"""All abstract classes later pipeline elements inherit from."""

import abc
import copy
import inspect
from collections.abc import Iterable
from typing import Any, NamedTuple, Self
from uuid import uuid4

from joblib import Parallel, delayed
from loguru import logger
from rdkit import Chem
from rdkit.Chem import Mol as RDKitMol
from sklearn.utils import Tags, TargetTags
from typing_extensions import override

from molpipeline.utils.multi_proc import check_available_cores


class InvalidInstance(NamedTuple):
    """Object which is returned when an instance cannot be processed.

    Attributes
    ----------
    element_id: str
        Id of the element which could not be processed.
    message: str
        Message why the element could not be processed.
    element_name: str | None
        Optional name of the element which could not be processed.
        The name of the pipeline element is often more descriptive than the id.

    """

    element_id: str
    message: str
    element_name: str | None = None

    def __repr__(self) -> str:
        """Return string representation of InvalidInstance.

        Returns
        -------
        str
            String representation of InvalidInstance.

        """
        return (
            f"InvalidInstance({self.element_name or self.element_id}, {self.message})"
        )


OptionalMol = RDKitMol | InvalidInstance


class RemovedInstance:  # pylint: disable=too-few-public-methods
    """Object which is returned by a ErrorFilter if an Invalid instance was removed."""

    def __init__(self, filter_element_id: str, message: str | None = None) -> None:
        """Initialize RemovedInstance.

        Parameters
        ----------
        filter_element_id: str
            FilterElement which removed the molecule.
        message: str | None, optional
            Information why the molecule was removed.

        """
        self.filter_element_id = filter_element_id
        self.message = message

    def __repr__(self) -> str:
        """Return string representation of RemovedInstance.

        Returns
        -------
        str
            String representation of RemovedInstance.

        """
        return f"RemovedInstance({self.filter_element_id}, {self.message})"


class ABCPipelineElement(abc.ABC):
    """Ancestor of all PipelineElements."""

    name: str
    _requires_fitting: bool = False
    uuid: str

    def __init__(
        self,
        name: str | None = None,
        n_jobs: int = 1,
        uuid: str | None = None,
    ) -> None:
        """Initialize ABCPipelineElement.

        Parameters
        ----------
        name: str | None, optional
            Name of PipelineElement
        n_jobs: int, default=1
            Number of cores used for processing.
        uuid: str | None, optional
            Unique identifier of the PipelineElement.

        """
        if name is None:
            name = self.__class__.__name__
        self.name = name
        self.n_jobs = n_jobs
        if uuid is None:
            self.uuid = str(uuid4())
        else:
            self.uuid = uuid

    def __repr__(self) -> str:
        """Return string representation of object.

        Returns
        -------
        str
            String representation of object.

        """
        parm_list = []
        for key, value in self._get_non_default_parameters().items():
            parm_list.append(f"{key}={value}")
        parm_str = ", ".join(parm_list)
        return f"{self.__class__.__name__}({parm_str})"

    def __sklearn_tags__(self) -> Tags:
        """Return Tags for the Element.

        Returns
        -------
        Tags
            Tags for the Element.

        """
        return Tags(
            estimator_type=None,
            target_tags=TargetTags(required=False),
            transformer_tags=None,
            regressor_tags=None,
            classifier_tags=None,
        )

    def _get_non_default_parameters(self) -> dict[str, Any]:
        """Return all parameters which are not default.

        Returns
        -------
        dict[str, Any]
            Dictionary of non default parameters.

        """
        signature = inspect.signature(self.__class__.__init__)
        self_params = self.get_params()
        non_default_params = {}
        for parm, value_default in signature.parameters.items():
            if parm == "uuid":
                continue
            if value_default is inspect.Parameter.empty:
                non_default_params[parm] = self_params[parm]
                continue
            if parm in self_params and value_default.default != self_params[parm]:
                non_default_params[parm] = self_params[parm]
        return non_default_params

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Return the parameters of the object.

        Parameters
        ----------
        deep: bool, default=True
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Parameters of the object.

        """
        if deep:
            return {
                "name": copy.copy(self.name),
                "n_jobs": copy.copy(self.n_jobs),
                "uuid": copy.copy(self.uuid),
            }

        return {
            "name": self.name,
            "n_jobs": self.n_jobs,
            "uuid": self.uuid,
        }

    def set_params(self, **parameters: Any) -> Self:
        """Change the parameters of the object.

        Parameters
        ----------
        parameters: Any
            Parameters to be set.

        Returns
        -------
        Self
            Self with updated parameters.

        Raises
        ------
        ValueError
            If the parameter is not a valid parameter of the object.

        """
        for att_name, att_value in parameters.items():
            if not hasattr(self, att_name):
                raise ValueError(
                    f"Cannot set attribute {att_name} on {self.__class__.__name__}",
                )
            setattr(self, att_name, att_value)
        return self

    @property
    def n_jobs(self) -> int:
        """Get the number of cores."""
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, n_jobs: int) -> None:
        """Set the number of cores.

        Parameters
        ----------
        n_jobs: int
            Number of cores used for processing.

        """
        self._n_jobs = check_available_cores(n_jobs)

    @property
    def requires_fitting(self) -> bool:
        """Return whether the object requires fitting or not."""
        return self._requires_fitting

    def fit(self, values: Any, labels: Any = None) -> Self:
        """Fit object to input_values.

        Most objects might not need fitting, but it is implemented for
        consistency for all PipelineElements.

        Parameters
        ----------
        values: Any
            List of molecule representations.
        labels: Any
            Optional label for fitting.

        Returns
        -------
        Self
            Fitted object.

        """
        _ = self.fit_transform(values, labels)
        return self

    def fit_to_result(self, values: Any) -> Self:  # noqa: ARG002
        """Fit object to result of transformed values.

        Fit object to the result of the transform function. This is useful for catching
        invalid or removed molecules.

        Parameters
        ----------
        values: Any
            List of molecule representations.

        Returns
        -------
        Self
            Fitted object.

        """
        return self

    @abc.abstractmethod
    def fit_transform(
        self,
        values: Any,
        labels: Any = None,
    ) -> Any:
        """Apply fit function and subsequently transform the input.

        Parameters
        ----------
        values: Any
            Apply transformation specified in transform_single to the value_list.
        labels: Any
            Optional label for fitting.

        Returns
        -------
        Any
            List of instances in new representation.

        """

    @abc.abstractmethod
    def transform(self, values: Any) -> Any:
        """Transform input_values according to object rules.

        Parameters
        ----------
        values: Any
            Iterable of instances to be transformed.
            Could be list of molecule representations (SMILES, Molecules, PhysChem
            vectors etc.).

        Returns
        -------
        Any
            Transformed input_values.

        """

    @abc.abstractmethod
    def transform_single(self, value: Any) -> Any:
        """Transform a single value.

        Parameters
        ----------
        value: Any
            Value to be transformed.

        Returns
        -------
        Any
            Transformed value.

        """


class TransformingPipelineElement(ABCPipelineElement):
    """Ancestor of all PipelineElements."""

    _input_type: str
    _output_type: str
    name: str

    def __init__(
        self,
        name: str | None = None,
        n_jobs: int = 1,
        uuid: str | None = None,
    ) -> None:
        """Initialize ABCPipelineElement.

        Parameters
        ----------
        name: str | None, optional
            Name of PipelineElement
        n_jobs: int, default=1
            Number of cores used for processing.
        uuid: str | None, optional
            Unique identifier of the PipelineElement.

        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        self._is_fitted = False

    @property
    def input_type(self) -> str:
        """Return the input type."""
        return self._input_type

    @property
    def is_fitted(self) -> bool:
        """Return whether the object is fitted or not."""
        return self._is_fitted

    @property
    def output_type(self) -> str:
        """Return the output type."""
        return self._output_type

    @property
    def parameters(self) -> dict[str, Any]:
        """Return the parameters of the object."""
        return self.get_params()

    @parameters.setter
    def parameters(self, parameters: Any) -> None:
        """Set the parameters of the object.

        Parameters
        ----------
        parameters: Any
            Object parameters as a dictionary.

        """
        self.set_params(**parameters)

    def fit_to_result(self, values: Any) -> Self:
        """Fit object to result of transformed values.

        Fit object to the result of the transform function. This is useful
        catching invalid or removed molecules.

        Parameters
        ----------
        values: Any
            List of molecule representations.

        Returns
        -------
        Self
            Fitted object.

        """
        self._is_fitted = True
        return super().fit_to_result(values)

    @override
    def fit_transform(
        self,
        values: Any,
        labels: Any = None,
    ) -> Any:
        """Apply fit function and subsequently transform the input.

        Parameters
        ----------
        values: Any
            Apply transformation specified in transform_single to the value_list.
        labels: Any
            Optional label for fitting.

        Returns
        -------
        Any
            List of molecules in new representation.

        """
        self._is_fitted = True
        if self.requires_fitting:
            pre_value_list = self.pretransform(values)
            self.fit_to_result(pre_value_list)
            output_list = self.finalize_list(pre_value_list)
            if hasattr(self, "assemble_output"):
                return self.assemble_output(output_list)
        return self.transform(values)

    def transform_single(self, value: Any) -> Any:
        """Transform a single molecule to the new representation.

        RemovedMolecule objects are passed without change.

        Parameters
        ----------
        value: Any
            Current representation of the molecule.

        Returns
        -------
        Any
            New representation of the molecule.

        """
        if isinstance(value, InvalidInstance):
            return value
        pre_value = self.pretransform_single(value)
        if isinstance(pre_value, InvalidInstance):
            return pre_value
        return self.finalize_single(pre_value)

    def assemble_output(self, value_list: Iterable[Any]) -> Any:
        """Aggregate rows, which in most cases is just return the list.

        Some representations might be better representd as a single object.
        For example a list of vectors can be transformed to a matrix.

        Parameters
        ----------
        value_list: Iterable[Any]
            Iterable of transformed rows.

        Returns
        -------
        Any
            Aggregated output. This can also be the original input.

        """
        return list(value_list)

    @abc.abstractmethod
    def pretransform_single(self, value: Any) -> Any:
        """Transform the instance, but skips parameters learned during fitting.

        This is the first step for the full transformation. It is followed by the
        finalize_single method and assemble output which collects all single
        transformations. These functions are split as they need to be accessed
        separately from outside the object.

        Parameters
        ----------
        value: Any
            Value to be pretransformed.

        Returns
        -------
        Any
            Pretransformed value.

        """

    def finalize_single(self, value: Any) -> Any:
        """Apply parameters learned during fitting to a single instance.

        Parameters
        ----------
        value: Any
            Value obtained from pretransform_single.

        Returns
        -------
        Any
            Finalized value.

        """
        return value

    def pretransform(self, value_list: Iterable[Any]) -> list[Any]:
        """Transform without using fitting specific parameters.

        Parameters
        ----------
        value_list: Iterable[Any]
            Iterable of instances to be pretransformed.

        Returns
        -------
        list[Any]
            Transformed input_values.

        """
        parallel = Parallel(n_jobs=self.n_jobs)
        return parallel(
            delayed(self.pretransform_single)(value) for value in value_list
        )

    def finalize_list(self, value_list: Iterable[Any]) -> list[Any]:
        """Transform list of values according to parameters learned during fitting.

        Parameters
        ----------
        value_list:  Iterable[Any]
            List of values to be transformed.

        Returns
        -------
        list[Any]
            List of transformed values.

        """
        parallel = Parallel(n_jobs=self.n_jobs)
        return parallel(delayed(self.finalize_single)(value) for value in value_list)

    def transform(self, values: Any) -> Any:
        """Transform input_values according to object rules.

        Parameters
        ----------
        values: Any
            Iterable of instances to be transformed.
            Could be list of molecule representations (SMILES, Molecules, PhysChem
            vectors etc.).

        Returns
        -------
        Any
            Transformed input_values.

        """
        output_rows = self.pretransform(values)
        output_rows = self.finalize_list(output_rows)
        return self.assemble_output(output_rows)


class MolToMolPipelineElement(TransformingPipelineElement, abc.ABC):
    """Abstract PipelineElement where input and outputs are molecules."""

    _input_type = "RDKitMol"
    _output_type = "RDKitMol"

    def transform(self, values: list[OptionalMol]) -> list[OptionalMol]:
        """Transform list of molecules to list of molecules.

        Parameters
        ----------
        values: list[OptionalMol]
            List of molecules to be transformed.

        Returns
        -------
        list[OptionalMol]
            List of molecules.
            InvalidInstances, if the transformation was not successful.

        """
        mol_list: list[OptionalMol] = super().transform(values)  # Stupid mypy...
        return mol_list

    def transform_single(self, value: OptionalMol) -> OptionalMol:
        """Apply pretransform_single and finalize_single in one step.

        Parameters
        ----------
        value: OptionalMol
            Molecule to be transformed.

        Returns
        -------
        OptionalMol
            Transformed molecule if transformation was successful, else InvalidInstance.

        """
        try:
            return super().transform_single(value)
        except Exception as exc:
            if isinstance(value, Chem.Mol):
                logger.error(
                    f"Failed to process: {Chem.MolToSmiles(value)} | "
                    f"({self.name}) ({self.uuid})",
                )
            else:
                logger.error(
                    f"Failed to process: {value} ({type(value)}) | "
                    f"({self.name}) ({self.uuid})",
                )
            raise exc

    @abc.abstractmethod
    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Transform the molecule to another molecule object.

        Do not apply parameters learned during fitting.

        Parameters
        ----------
        value: RDKitMol
            Molecule to be transformed.

        Returns
        -------
        OptionalMol
            Transformed molecule if transformation was successful, else InvalidInstance.

        """


class AnyToMolPipelineElement(TransformingPipelineElement, abc.ABC):
    """Abstract PipelineElement which creates molecules from different inputs."""

    _output_type = "RDKitMol"

    def transform(self, values: Any) -> list[OptionalMol]:
        """Transform list of instances to list of molecules.

        Parameters
        ----------
        values: Any
            Instances to be transformed to a list of molecules.

        Returns
        -------
        list[OptionalMol]
            List of molecules.
            InvalidInstance if the representation was invalid.

        """
        mol_list: list[OptionalMol] = super().transform(values)  # Stupid mypy...
        return mol_list

    @abc.abstractmethod
    def pretransform_single(self, value: Any) -> OptionalMol:
        """Transform a single instance to a molecule.

        The transformations only includes steps which do not depend on parameters
        learned during fitting.

        Parameters
        ----------
        value: Any
            Representation to be transformed to a molecule.

        Returns
        -------
        OptionalMol
            Obtained molecule if valid representation, else InvalidInstance.

        """


class MolToAnyPipelineElement(TransformingPipelineElement, abc.ABC):
    """Abstract PipelineElement which creates molecules from different inputs."""

    _input_type = "RDKitMol"

    @abc.abstractmethod
    def pretransform_single(self, value: RDKitMol) -> Any:
        """Transform the molecule, but skip parameters learned during fitting.

        Parameters
        ----------
        value: RDKitMOl
            Molecule to be transformed.

        Returns
        -------
        Any
            Transformed molecule.

        """



"""Initialize module for file loading."""

import importlib.util

from molpipeline.utils.file_loading.dvc_file_loading import ABCFileLoader

__all__ = [
    "ABCFileLoader",
]

if importlib.util.find_spec("requests") is not None:
    from molpipeline.utils.file_loading.url_file_loading import URLFileLoader

    __all__ += ["URLFileLoader"]


if importlib.util.find_spec("dvc") is not None:
    from molpipeline.utils.file_loading.dvc_file_loading import DVCFileLoader

    __all__ += ["DVCFileLoader"]



"""Unit tests for BootstrapSplit splitter."""

import unittest

import numpy as np

from molpipeline.experimental.model_selection.splitter.bootstrap_splitter import (
    BootstrapSplit,
)


class TestBootstrapSplit(unittest.TestCase):
    """Tests for BootstrapSplit."""

    def test_invalid_max_samples(self) -> None:
        """Verify that max_samples cannot be set with invalid value."""
        invalid_values = [-0.5, 0.0, 1.5]
        for value in invalid_values:
            with (
                self.subTest(max_samples=value),
                self.assertRaisesRegex(
                    ValueError,
                    r"If max_samples is a float, it must be in the range \(0.0, 1.0].",
                ),
            ):
                BootstrapSplit(n_splits=3, max_samples=1.5)

    def test_get_n_splits(self) -> None:
        """Verify get_n_splits returns the configured number of splits."""
        splitter = BootstrapSplit(n_splits=4, random_state=0)
        self.assertEqual(splitter.get_n_splits(X=np.ones(3)), 4)

    def test_split_properties(self) -> None:
        """Verify train/test indices satisfy bootstrap properties."""
        n_samples = 20
        n_splits = 5
        x = np.arange(n_samples)
        splitter = BootstrapSplit(n_splits=n_splits, random_state=7)

        splits = list(splitter.split(X=x))
        self.assertEqual(len(splits), n_splits)

        all_indices = np.arange(n_samples)
        for train_indices, test_indices in splits:
            # Train set is sampled with replacement and has same size as input.
            self.assertEqual(train_indices.shape[0], n_samples)

            # Test set is exactly the complement of sampled train indices.
            expected_test = np.setdiff1d(all_indices, train_indices)
            self.assertGreater(len(expected_test), 0)  # Test set is non-empty
            self.assertTrue(np.array_equal(test_indices, expected_test))
            # Test that test_indices does not contain repetitions
            self.assertEqual(test_indices.shape[0], np.unique(test_indices).shape[0])

            # Test all samples are covered by either train or test sets.
            unique_train = np.unique(train_indices)
            unique_test = np.unique(test_indices)
            self.assertEqual(np.union1d(unique_train, unique_test).shape[0], n_samples)

    def test_split_reproducible_and_expected_for_seed(self) -> None:
        """Verify split output is deterministic for a fixed random_state."""
        x = np.arange(6)
        n_splits = 5
        splitter_a = BootstrapSplit(n_splits=n_splits, random_state=123)
        splitter_b = BootstrapSplit(n_splits=n_splits, random_state=123)
        splits_a = list(splitter_a.split(X=x))
        splits_b = list(splitter_b.split(X=x))

        self.assertEqual(len(splits_a), n_splits)
        self.assertEqual(len(splits_b), n_splits)

        self.assertEqual(len(splits_a), len(splits_b))
        for split_a_i, split_b_i in zip(splits_a, splits_b, strict=True):
            train_a_i, test_a_i = split_a_i
            train_b_i, test_b_i = split_b_i
            self.assertTrue(np.array_equal(train_a_i, train_b_i))
            self.assertTrue(np.array_equal(test_a_i, test_b_i))

    def test_no_random_state(self) -> None:
        """Verify that a None is a valid random state."""
        n_splits = 5
        splitter = BootstrapSplit(n_splits=n_splits, random_state=None)
        splits = list(splitter.split(np.ones(30)))
        self.assertEqual(len(splits), n_splits)

    def test_split_with_integer_max_samples(self) -> None:
        """Verify integer max_samples controls sampled population and draw size."""
        max_samples = 7
        splitter = BootstrapSplit(n_splits=3, max_samples=max_samples, random_state=17)

        for train_indices, _test_indices in splitter.split(X=np.arange(20)):
            self.assertEqual(train_indices.shape[0], max_samples)

        sample_size = 5
        for train_indices, _test_indices in splitter.split(X=np.arange(sample_size)):
            self.assertEqual(train_indices.shape[0], sample_size)

    def test_split_with_float_max_samples(self) -> None:
        """Verify float max_samples uses the corresponding fraction of samples."""
        x = np.arange(25)
        max_samples = 0.4
        expected_n_samples = int(len(x) * max_samples)
        splitter = BootstrapSplit(n_splits=4, max_samples=max_samples, random_state=9)
        for train_indices, _test_indices in splitter.split(X=x):
            self.assertEqual(train_indices.shape[0], expected_n_samples)


if __name__ == "__main__":
    unittest.main()



"""Utilities for randomization ensuring that tests do not rely on specific values."""

import numpy as np
import torch


def randomize_state_dict_weights(
    state_dict: dict[str, torch.Tensor],
    random_state: int | None = None,
) -> dict[str, torch.Tensor]:
    """Randomize the weights in a state_dict.

    Parameters
    ----------
    state_dict : dict[str, torch.Tensor]
        The state_dict to randomize.
    random_state : int | None, default=None
        The random state to use for reproducibility.

    Returns
    -------
    dict[str, torch.Tensor]
        The randomized state_dict.

    Raises
    ------
    ValueError
        If the state dict does not contain any weights to randomize.

    """
    random_state_dict = {}
    rng = np.random.default_rng(random_state)
    altered_keys = []
    for key, value in state_dict.items():
        if key.endswith(".weight"):
            altered_keys.append(key)
            random_weights = rng.normal(size=value.shape).astype(np.float32)
            random_state_dict[key] = torch.tensor(random_weights)
        else:
            random_state_dict[key] = value
    if not altered_keys:
        raise ValueError("No weights were altered in the state_dict.")
    return random_state_dict



"""Conformal prediction subpackage for classification and regression.

This package provides conformal prediction wrappers using crepes:
- ConformalClassifier / CrossConformalClassifier
- ConformalRegressor / CrossConformalRegressor
- BaseConformalPredictor
- Nonconformity utilities (via uncertainty.utils)
"""

from molpipeline.experimental.uncertainty.conformal.conformal_base import (
    BaseConformalPredictor,
    _apply_antitonic_regressors,
    _fit_antitonic_regressors,
)
from molpipeline.experimental.uncertainty.conformal.conformal_classsifier import (
    ConformalClassifier,
    CrossConformalClassifier,
)
from molpipeline.experimental.uncertainty.conformal.conformal_regressor import (
    ConformalRegressor,
    CrossConformalRegressor,
)

__all__ = [
    "BaseConformalPredictor",
    "ConformalClassifier",
    "ConformalRegressor",
    "CrossConformalClassifier",
    "CrossConformalRegressor",
    "_apply_antitonic_regressors",
    "_fit_antitonic_regressors",
]



"""Tests conversion of sklearn models to json and back."""

import unittest
from typing import Any

import numpy as np
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import balanced_accuracy_score
from sklearn.model_selection import StratifiedKFold
from sklearn.svm import SVC

from molpipeline import Pipeline
from molpipeline.utils.json_operations import (
    get_init_params,
    recursive_from_json,
    recursive_to_json,
)


class JsonConversionTest(unittest.TestCase):
    """Unittest for conversion of sklearn models to json and back."""

    def test_rf_reconstruction(self) -> None:
        """Test if the sklearn-rf can be reconstructed from json."""
        random_forest = RandomForestClassifier(n_estimators=200)
        recreated_rf = recursive_from_json(recursive_to_json(random_forest))
        self.assertEqual(random_forest.get_params(), recreated_rf.get_params())

    def test_svc_reconstruction(self) -> None:
        """Test if the sklearn-svc can be reconstructed from json."""
        svc = SVC()
        recreated_svc = recursive_from_json(recursive_to_json(svc))
        self.assertEqual(svc.get_params(), recreated_svc.get_params())

    def test_pipeline_reconstruction(self) -> None:
        """Test if the sklearn-pipleine can be reconstructed from json."""
        random_forest = RandomForestClassifier(n_estimators=200)
        svc = SVC()
        pipeline = Pipeline([("rf", random_forest), ("svc", svc)])
        recreated_pipeline = recursive_from_json(recursive_to_json(pipeline))

        original_params = pipeline.get_params()
        recreated_params = recreated_pipeline.get_params()
        original_steps = original_params.pop("steps")
        recreated_steps = recreated_params.pop("steps")

        # Separate comparison of the steps as models cannot be compared directly
        for (orig_name, orig_obj), (recreated_name, recreated_obj) in zip(
            original_steps,
            recreated_steps,
            strict=True,
        ):
            # Remove the model from the original params
            del original_params[orig_name]
            del recreated_params[recreated_name]
            self.assertEqual(orig_name, recreated_name)
            self.assertEqual(orig_obj.get_params(), recreated_obj.get_params())
            self.assertEqual(type(orig_obj), type(recreated_obj))
        self.assertEqual(original_params, recreated_params)

    def test_set_transformation(self) -> None:
        """Test if a set can be reconstructed from json."""
        test_set = {1, "a", (1, "a")}
        test_set_json = recursive_to_json(test_set)
        recreated_set = recursive_from_json(test_set_json)
        self.assertEqual(test_set, recreated_set)

    def test_numpy_dtype_roundtrip(self) -> None:
        """Test if a numpy dtype can be reconstructed from json."""
        dtype_dict = {
            "int64": np.int64,
            "float32": np.float32,
            "object": np.object_,
        }
        for dtype_name, dtype in dtype_dict.items():
            with self.subTest(dtype=dtype_name):
                json_data = recursive_to_json(dtype)
                deserialized_dtype = recursive_from_json(json_data)
                self.assertEqual(dtype, deserialized_dtype)

    def test_array_roundtrip(self) -> None:
        """Test if a numpy array can be reconstructed from json."""
        dtype_dict = {
            "int64": np.int64,
            "float32": np.float32,
            "object": np.object_,
        }
        for dtype_name, dtype in dtype_dict.items():
            with self.subTest(dtype=dtype_name):
                original_array = np.array([[1, 2], [3, 4]], dtype=dtype)
                json_data = recursive_to_json(original_array)
                deserialized_array = recursive_from_json(json_data)
                self.assertTrue(np.array_equal(original_array, deserialized_array))
                self.assertEqual(original_array.dtype, deserialized_array.dtype)

    def test_function_roundtrip(self) -> None:
        """Test if a function can be reconstructed from json."""
        json_data = recursive_to_json(balanced_accuracy_score)
        deserialized_function = recursive_from_json(json_data)
        self.assertIs(deserialized_function, balanced_accuracy_score)


class ReconstructibleObject:  # pylint: disable=too-few-public-methods
    """Simple class with stable reconstruction state."""

    def __init__(self, required: int, optional: int = 2) -> None:
        """Initialize the object.

        Parameters
        ----------
        required : int
            A required parameter.
        optional : int, default=2
            An optional parameter with a default value.

        """
        self.required = required
        self.optional = optional
        self.sum = required + optional


class MissingRequiredStateParamObj(ReconstructibleObject):  # pylint: disable=too-few-public-methods
    """Class that omits a required init parameter from state."""

    def __getstate__(self) -> dict[str, Any]:
        """Mock the get_state and omit the parameter `required`.

        Returns
        -------
        dict[str, Any]
            State dictionary missing the required parameter.

        """
        state_dict = dict(self.__dict__)
        state_dict.pop("required")  # Remove the required parameter
        return state_dict


class ModifiedParamObj(ReconstructibleObject):  # pylint: disable=too-few-public-methods
    """Class that modifies a required init parameter in state."""

    def __init__(self, required: int, optional: int = 2) -> None:
        """Initialize the object.

        Parameters
        ----------
        required : int
            A required parameter.
        optional : int, default=2
            An optional parameter with a default value.

        """
        super().__init__(required * 2, optional)


class GetInitParamsTest(unittest.TestCase):
    """Test the get_init_params function."""

    def test_estimator(self) -> None:
        """Test extracting init params from a sklearn estimator."""
        estimator = RandomForestClassifier(n_estimators=15, random_state=7)
        init_params = get_init_params(estimator, validation="raise")
        self.assertEqual(init_params, estimator.get_params(deep=False))

    def test_splitter(self) -> None:
        """Test extracting init params from a sklearn splitter."""
        orig_params = {"n_splits": 4, "shuffle": True, "random_state": 13}
        splitter = StratifiedKFold(**orig_params)
        init_params = get_init_params(splitter, validation="raise")
        self.assertEqual(init_params, orig_params)

    def test_non_sklearn_object(self) -> None:
        """Test extracting init params from a custom serializable class."""
        custom_obj = ReconstructibleObject(required=3, optional=5)
        init_params = get_init_params(custom_obj, validation="raise")
        self.assertEqual(init_params, {"required": 3, "optional": 5})

    def test_return_none_for_missing_required_params(self) -> None:
        """Test return_none validation strategy for missing required params."""
        custom_obj = MissingRequiredStateParamObj(required=5, optional=9)
        init_params = get_init_params(custom_obj, validation="return_none")
        self.assertIsNone(init_params)

    def test_raises_for_missing_required_params(self) -> None:
        """Test raise validation strategy for missing required params."""
        custom_obj = MissingRequiredStateParamObj(required=5, optional=9)
        with self.assertRaises(ValueError):
            get_init_params(custom_obj, validation="raise")

    def test_missing_required_params_skip(self) -> None:
        """Test skip validation strategy for missing required params."""
        custom_obj = MissingRequiredStateParamObj(required=5, optional=9)
        init_params = get_init_params(custom_obj, validation="skip")
        self.assertIsNotNone(init_params)
        self.assertEqual(init_params, {"optional": 9})

    def test_modified_param_obj(self) -> None:
        """Test that modified parameters raise an error."""
        custom_obj = ModifiedParamObj(required=4, optional=6)
        with self.assertRaisesRegex(ValueError, "Reconstruction of the object failed."):
            get_init_params(custom_obj, validation="raise")


if __name__ == "__main__":
    unittest.main()



"""Implementation for 2D pharmacophore fingerprints."""

import copy
from pathlib import Path
from typing import Any, Literal, Self

import numpy as np
import numpy.typing as npt
from rdkit.Chem.Pharm2D import Gobbi_Pharm2D
from rdkit.Chem.Pharm2D.Generate import Gen2DFingerprint
from rdkit.Chem.Pharm2D.SigFactory import SigFactory
from rdkit.Chem.rdMolChemicalFeatures import (
    BuildFeatureFactoryFromString,
)
from rdkit.DataStructs import ConvertToExplicit, ExplicitBitVect, IntSparseIntVect
from rdkit.RDPaths import RDDataDir

from molpipeline.abstract_pipeline_elements.mol2any.mol2bitvector import (
    FPReturnAsOption,
    MolToFingerprintPipelineElement,
)
from molpipeline.utils.molpipeline_types import RDKitMol

_MIN_POINT_COUNT = 2


class MolToPharmacophore2DFP(  # pylint: disable=too-many-instance-attributes
    MolToFingerprintPipelineElement,
):
    """2D Pharmacophore Fingerprint.

    Computes RDKit's 2D pharmacophore fingerprints based on pharmacophore features
    and their 2D distances. The fingerprint encodes the presence of pharmacophore
    feature pairs at specific distance ranges. Distances are determined on the
    molecular graph, not in 3D space.

    This implementation uses RDKit's Pharm2D module which generates fingerprints
    based on:
    - Feature definitions (e.g., donors, acceptors, aromatic rings)
    - Distance bins for feature pairs
    - Configurable parameters for feature factory and signature factory

    Per default, the 2d pharmacophore fingerprint described by Gobbi et al. is used.

    References
    ----------
    Gobbi, A. & Poppinger, D. Genetic optimization of combinatorial libraries.
        Biotechnology and Bioengineering 61, 47-54 (1998).

    """

    def __init__(
        self,
        *,
        feature_definition: str | None = None,
        min_point_count: int = 2,
        max_point_count: int = 3,
        triangular_pruning: bool = True,
        shortest_paths_only: bool = True,
        include_bond_order: bool = False,
        skip_feats: list[str] | None = None,
        distance_bins: list[tuple[float, float]] | None = None,
        counted: bool = False,
        return_as: FPReturnAsOption = "sparse",
        name: str = "MolToPharmacophore2DFP",
        n_jobs: int = 1,
        uuid: str | None = None,
    ) -> None:
        """Initialize MolToPharmacophore2DFP.

        Parameters
        ----------
        feature_definition : str | None, optional
            Path or content of a feature definition file (.fdef). If None, uses
            configuration by Gobbi et al.
        min_point_count : int, default=2
            Minimum number of pharmacophore points in a signature.
        max_point_count : int, default=3
            Maximum number of pharmacophore points in a signature.
        triangular_pruning : bool, default=True
            Whether to use triangular pruning for efficiency.
        shortest_paths_only : bool, default=True
            Whether to use only shortest paths between features.
        include_bond_order : bool, default=False
            Whether to include bond order information in the fingerprint.
        skip_feats : list[str], optional
            List of feature types to skip. If None, no features are skipped.
        distance_bins : list[tuple[float, float]], optional
            List of distance bins as (min_distance, max_distance) tuples.
            If None, uses default bins by Gobbi et al.
        counted : bool, default=False
            If True, the fingerprint will be counted (values represent occurrence).
            If False, the fingerprint will be binary (values are 0 or 1).
        return_as : Literal["sparse", "dense", "rdkit"], default="sparse"
            Type of output. When "sparse" the fingerprints will be returned as a
            scipy.sparse.csr_matrix holding a sparse representation of the bit vectors.
            With "dense" a numpy matrix will be returned.
            With "rdkit" the fingerprints will be returned as a list of
            RDKit's internal data structure. ExplicitBitVect if counted is False,
            IntSparseIntVect if counted is True.
        name : str, default="MolToPharmacophore2DFP"
            Name of PipelineElement.
        n_jobs : int, default=1
            Number of cores to use.
        uuid : str, optional
            UUID of the PipelineElement.

        """
        super().__init__(  # pylint: disable=R0801
            return_as=return_as,
            name=name,
            n_jobs=n_jobs,
            uuid=uuid,
        )

        self._validate_min_max_point_count(min_point_count, max_point_count)

        self._min_point_count = min_point_count
        self._max_point_count = max_point_count
        self._triangular_pruning = triangular_pruning
        self._shortest_paths_only = shortest_paths_only
        self._include_bond_order = include_bond_order
        self._skip_feats = skip_feats if skip_feats is not None else []
        self._counted = counted
        self._feature_definition = feature_definition or Gobbi_Pharm2D.fdef
        self._distance_bins = distance_bins or Gobbi_Pharm2D.defaultBins

        self._validate_distance_bins(self._distance_bins)

        # Initialize factories and calculate fingerprint size
        self._initialize_factories()

    def _initialize_factories(self) -> None:
        """Initialize the feature factory and signature factory.

        Raises
        ------
        ValueError
            If the signature factory does not produce any bits.

        """
        feature_factory = BuildFeatureFactoryFromString(self._feature_definition)

        # Create signature factory
        self._sig_factory = SigFactory(
            feature_factory,
            minPointCount=self._min_point_count,
            maxPointCount=self._max_point_count,
            trianglePruneBins=self._triangular_pruning,
            shortestPathsOnly=self._shortest_paths_only,
            includeBondOrder=self._include_bond_order,
            skipFeats=self._skip_feats,
            useCounts=self._counted,
        )

        # Set distance bins
        self._sig_factory.SetBins(self._distance_bins)

        # Initialize the signature factory to get the fingerprint size
        self._sig_factory.Init()
        self._n_bits = self._sig_factory.GetSigSize()

        if self._n_bits == 0:
            raise ValueError(
                "The signature factory did not produce any bits. "
                "Check the feature definition and parameters.",
            )

        # Generate feature names
        self._feature_names = [
            f"pharm2d_{self._sig_factory.GetBitDescription(i)}"
            for i in range(self._n_bits)
        ]

    def __getstate__(self) -> dict[str, Any]:
        """Get the state of the object for pickling.

        Returns
        -------
        dict[str, Any]
            Dictionary containing the state of the object.

        """
        state = self.__dict__.copy()
        # Remove the SigFactory from the state to avoid pickling issues
        state.pop("_sig_factory", None)
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Set the state of the object after unpickling.

        Parameters
        ----------
        state : dict[str, Any]
            Dictionary containing the state of the object.

        """
        self.__dict__.update(state)
        # Reinitialize the signature factory after unpickling
        self._initialize_factories()

    @staticmethod
    def _validate_min_max_point_count(
        min_point_count: int,
        max_point_count: int,
    ) -> None:
        """Validate the minimum and maximum point count.

        Parameters
        ----------
        min_point_count : int
            Minimum point count.
        max_point_count : int
            Maximum point count.

        Raises
        ------
        ValueError
            If min_point_count is less than 2 or greater than max_point_count.
            If max_point_count is less than 2.

        """
        if min_point_count < _MIN_POINT_COUNT:
            raise ValueError(
                f"Minimum point count must be at least {_MIN_POINT_COUNT}.",
            )
        if min_point_count > max_point_count:
            raise ValueError(
                "Minimum point count cannot be greater than maximum point count.",
            )

    @staticmethod
    def _validate_distance_bins(
        distance_bins: list[tuple[float, float]] | None,
    ) -> None:
        """Validate distance bins format and values.

        Parameters
        ----------
        distance_bins : list[tuple[float, float]] | None
            List of distance bins as (min_distance, max_distance) tuples.

        Raises
        ------
        ValueError
            If any distance bin does not contain numeric values,
            if min_distance is not less than max_distance,
            or if min_distance is negative.

        """
        if distance_bins is None:
            return
        for i, (min_dist, max_dist) in enumerate(distance_bins):
            if not isinstance(min_dist, (int, float)) or not isinstance(
                max_dist,
                (int, float),
            ):
                raise ValueError(f"Distance bin {i} must contain numeric values")
            if min_dist >= max_dist:
                raise ValueError(
                    f"Distance bin {i}: min_dist must be less than max_dist",
                )
            if min_dist < 0:
                raise ValueError(f"Distance bin {i}: distances must be non-negative")

    @property
    def min_point_count(self) -> int:
        """Get minimum point count."""
        return self._min_point_count

    @property
    def max_point_count(self) -> int:
        """Get maximum point count."""
        return self._max_point_count

    @property
    def triangular_pruning(self) -> bool:
        """Get triangular pruning setting."""
        return self._triangular_pruning

    @property
    def shortest_paths_only(self) -> bool:
        """Get shortest paths only setting."""
        return self._shortest_paths_only

    @property
    def include_bond_order(self) -> bool:
        """Get include bond order setting."""
        return self._include_bond_order

    @property
    def skip_feats(self) -> list[str]:
        """Get list of features to skip."""
        return self._skip_feats[:]

    @property
    def feature_definition(self) -> str:
        """Get feature factory file path."""
        return self._feature_definition

    @property
    def distance_bins(self) -> list[tuple[float, float]]:
        """Get distance bins."""
        return self._distance_bins[:]

    @property
    def counted(self) -> bool:
        """Get counted setting."""
        return self._counted

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Return all parameters defining the object.

        Parameters
        ----------
        deep : bool, default=True
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary of parameters.

        """
        parameters = super().get_params(deep)
        if deep:
            parameters.update(
                {
                    "feature_definition": copy.copy(self._feature_definition),
                    "min_point_count": copy.copy(self._min_point_count),
                    "max_point_count": copy.copy(self._max_point_count),
                    "triangular_pruning": copy.copy(self._triangular_pruning),
                    "shortest_paths_only": copy.copy(self._shortest_paths_only),
                    "include_bond_order": copy.copy(self._include_bond_order),
                    "skip_feats": copy.deepcopy(self._skip_feats),
                    "distance_bins": copy.deepcopy(self._distance_bins),
                    "counted": copy.copy(self._counted),
                },
            )
        else:
            parameters.update(
                {
                    "feature_definition": self._feature_definition,
                    "min_point_count": self._min_point_count,
                    "max_point_count": self._max_point_count,
                    "triangular_pruning": self._triangular_pruning,
                    "shortest_paths_only": self._shortest_paths_only,
                    "include_bond_order": self._include_bond_order,
                    "skip_feats": self._skip_feats,
                    "distance_bins": self._distance_bins,
                    "counted": self._counted,
                },
            )
        return parameters

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters.

        Parameters
        ----------
        parameters : Any
            Dictionary of parameter names and values.

        Returns
        -------
        Self
            MolToPharmacophore2DFP pipeline element with updated parameters.

        """
        parameter_copy = dict(parameters)

        # Set parameters that require reinitialization
        needs_reinit = False

        if "feature_definition" in parameter_copy:
            self._feature_definition = parameter_copy.pop("feature_definition")
            needs_reinit = True

        if "min_point_count" in parameter_copy:
            min_point_count = parameter_copy.pop("min_point_count")
            self._validate_min_max_point_count(min_point_count, self._max_point_count)
            self._min_point_count = min_point_count
            needs_reinit = True

        if "max_point_count" in parameter_copy:
            max_point_count = parameter_copy.pop("max_point_count")
            self._validate_min_max_point_count(self._min_point_count, max_point_count)
            self._max_point_count = max_point_count
            needs_reinit = True

        if "triangular_pruning" in parameter_copy:
            self._triangular_pruning = parameter_copy.pop("triangular_pruning")
            needs_reinit = True

        if "shortest_paths_only" in parameter_copy:
            self._shortest_paths_only = parameter_copy.pop("shortest_paths_only")
            needs_reinit = True

        if "include_bond_order" in parameter_copy:
            self._include_bond_order = parameter_copy.pop("include_bond_order")
            needs_reinit = True

        if "skip_feats" in parameter_copy:
            self._skip_feats = parameter_copy.pop("skip_feats")
            needs_reinit = True

        if "distance_bins" in parameter_copy:
            distance_bins = parameter_copy.pop("distance_bins")
            self._validate_distance_bins(distance_bins)
            self._distance_bins = distance_bins
            needs_reinit = True

        if "counted" in parameter_copy:
            self._counted = parameter_copy.pop("counted")
            needs_reinit = True

        # Reinitialize factories if needed
        if needs_reinit:
            self._initialize_factories()

        # Set remaining parameters via parent class
        super().set_params(**parameter_copy)

        return self

    def pretransform_single(
        self,
        value: RDKitMol,
    ) -> ExplicitBitVect | IntSparseIntVect | npt.NDArray[np.int_] | dict[int, int]:
        """Transform a single compound to a pharmacophore fingerprint.

        Parameters
        ----------
        value : RDKitMol
            Molecule for which the pharmacophore fingerprint is generated.

        Returns
        -------
        ExplicitBitVect | IntSparseIntVect | npt.NDArray[np.int_] | dict[int, int]
            If return_as is "rdkit" return ExplicitBitVect when counted=False, else
            IntSparseIntVect.
            If return_as is "dense" return numpy array.
            If return_as is "sparse" return dictionary with feature-position as key
            and count as value.

        """
        # fp will be of type `SparseBitVect` when counted is False, otherwise
        # `IntSparseIntVect`.
        fp = Gen2DFingerprint(value, self._sig_factory)
        if self._return_as == "dense":
            return np.array(fp.ToList())
        if self._return_as == "rdkit":
            if self.counted:
                return fp  # return IntSparseIntVect
            # Convert SparseBitVect to ExplicitBitVect
            return ConvertToExplicit(fp)

        # return sparse representation as dictionary
        if self.counted:
            return fp.GetNonzeroElements()
        return dict.fromkeys(fp.GetOnBits(), 1)

    @classmethod
    def from_file(cls, path: Path | str, **kwargs: Any) -> Self:
        """Create a MolToPharmacophore2DFP instance from a feature definition file.

        Parameters
        ----------
        path : Path | str
            Path to the feature definition file (.fdef).
        **kwargs : Any
            Additional parameters to the MolToPharmacophore2DFP constructor.

        Returns
        -------
        MolToPharmacophore2DFP
            Instance of MolToPharmacophore2DFP with the specified feature definition.

        Raises
        ------
        TypeError
            If the path is not a Path or str.

        """
        if not isinstance(path, (Path, str)):
            raise TypeError("path must be a Path or str.")
        return cls(feature_definition=Path(path).read_text(encoding="utf-8"), **kwargs)

    @classmethod
    def from_preconfiguration(
        cls,
        config_name: Literal["base", "gobbi"],
        **kwargs: Any,
    ) -> Self:
        """Create a preconfigured MolToPharmacophore2DFP instance.

        Preconfigurations:
        - "gobbi": Uses Gobbi's pharmacophore features as defined in:
           Gobbi, A. & Poppinger, D. Genetic optimization of combinatorial libraries.
           Biotechnology and Bioengineering 61, 47-54 (1998).
        - "base": Base configuration defined in RDKit.
           Warning: While this configuration is directly available in RDKit, it is
           unclear how stable it is and how well it was evaluated.
           See Greg's talk from RDKIT UGM 2012 slide 17
           https://www.rdkit.org/UGM/2012/Landrum_RDKit_UGM.Fingerprints.Final.pptx.pdf


        Parameters
        ----------
        config_name : Literal["gobbi"]
            Name of the preconfiguration to use.
        **kwargs : Any
            Additional parameters to the MolToPharmacophore2DFP constructor.

        Returns
        -------
        MolToPharmacophore2DFP
            Preconfigured MolToPharmacophore2DFP instance.

        Raises
        ------
        ValueError
            If the configuration name is unknown.

        """
        if config_name == "gobbi":
            # gobbi pharmacophore features are also implemented in RDKit. We just
            # borrow the definition here from the Gobbi_Pharm2D module.
            return cls(
                feature_definition=Gobbi_Pharm2D.fdef,
                min_point_count=2,
                max_point_count=3,
                distance_bins=Gobbi_Pharm2D.defaultBins,
                **kwargs,
            )
        if config_name == "base":
            # RDKit's "base" configuration
            # Feature definitions are taken from RDKit's BaseFeatures.fdef file.
            # Bins are taken from:
            # https://www.rdkit.org/UGM/2012/Landrum_RDKit_UGM.Fingerprints.Final.pptx.pdf
            return cls.from_file(
                Path(RDDataDir) / "BaseFeatures.fdef",
                distance_bins=[
                    (2, 3),
                    (3, 4),
                    (4, 5),
                    (5, 6),
                    (6, 7),
                    (7, 8),
                    (8, 100),
                ],
                **kwargs,
            )

        raise ValueError(f"Unknown configuration name: {config_name}")



"""Test module for behavior of chemprop in a pipeline."""

import unittest
from io import BytesIO
from typing import TypeVar
from unittest import mock

import joblib
import numpy as np
import pandas as pd
import torch
from sklearn.base import clone
from sklearn.calibration import CalibratedClassifierCV

from molpipeline.any2mol import SmilesToMol
from molpipeline.error_handling import ErrorFilter, FilterReinserter
from molpipeline.estimators.chemprop.abstract import ABCChemprop
from molpipeline.estimators.chemprop.component_wrapper import BondMessagePassing
from molpipeline.estimators.chemprop.metric_wrapper import BCELoss
from molpipeline.experimental.estimators.ensemble.homogeneous_ensemble import (
    HomogeneousEnsembleClassifier,
)
from molpipeline.mol2any.mol2chemprop import MolToChemprop
from molpipeline.pipeline import Pipeline
from molpipeline.post_prediction import PostPredictionWrapper
from molpipeline.utils.file_loading.url_file_loading import URLFileLoader
from molpipeline.utils.json_operations import recursive_from_json, recursive_to_json
from test_extras.test_chemprop.chemprop_test_utils.compare_models import compare_params
from test_extras.test_chemprop.chemprop_test_utils.default_models import (
    get_classification_pipeline,
    get_model,
    get_model_pipeline,
    get_multiclass_classification_pipeline,
    get_regression_pipeline,
    get_smiles_checker_pipeline,
)
from test_extras.test_chemprop.chemprop_test_utils.randomization import (
    randomize_state_dict_weights,
)
from tests import TEST_DATA_DIR

_T = TypeVar("_T")


def joblib_dump_load(obj: _T) -> _T:
    """Dump and load an object using joblib.

    Parameters
    ----------
    obj : _T
        The object to dump and load.

    Returns
    -------
    _T
        The loaded object.

    Notes
    -----
    The object is not dumped to disk but to a BytesIO object.

    """
    bytes_container = BytesIO()
    joblib.dump(obj, bytes_container)
    bytes_container.seek(0)  # update to enable reading
    bytes_model = bytes_container.read()

    return joblib.load(BytesIO(bytes_model))


class TestChempropPipeline(unittest.TestCase):
    """Test the Chemprop model pipeline."""

    def test_get_set_params(self) -> None:
        """Test the get_params and set_params methods of the pipeline."""
        model = get_model_pipeline()
        orig_params = model.get_params(deep=True)
        new_params = {
            "model__model__predictor__activation": "relu",
            "model__model__predictor__dropout": 0.5,
            "model__model__message_passing__activation": "relu",
        }
        # Check setting new parameters
        model.set_params(**new_params)
        model_params = model.get_params(deep=True)
        for param_name, param in new_params.items():
            self.assertEqual(param, model_params[param_name])
        model.set_params(**orig_params)
        model_params = model.get_params(deep=True)
        for param_name, param in orig_params.items():
            self.assertEqual(param, model_params[param_name])

    def test_clone(self) -> None:
        """Test the clone method of the pipeline."""
        model = get_model_pipeline()
        cloned_model = clone(model)
        for step_name, step in model.steps:
            cloned_step = cloned_model.named_steps[step_name]
            self.assertEqual(step.__class__, cloned_step.__class__)
            params = step.get_params(deep=True)  # type: ignore
            cloned_params = cloned_step.get_params(deep=True)
            if isinstance(step, ABCChemprop):
                compare_params(self, step, cloned_step)
                continue
            for param_name, param in params.items():
                # If parm implements get_params, it was cloned as well and we need to
                # compare the parameters. Since all parameters are listed flat in the
                # params dicts, all objects are identical if param dicts are identical.
                if hasattr(param, "get_params"):
                    self.assertEqual(
                        param.__class__,
                        cloned_params[param_name].__class__,
                    )
                else:
                    self.assertEqual(
                        param,
                        cloned_params[param_name],
                        f"Failed for {param_name}",
                    )

    def test_passing_smiles(self) -> None:
        """Test passing SMILES strings to the pipeline.

        Since the pipeline is not trained, we only check if the prediction is
        successful.

        """
        model = get_model_pipeline()
        smiles = ["CCO", "CNC", "CCN"]
        pred = model.predict(smiles)
        self.assertEqual(len(pred), 3)
        proba = model.predict_proba(smiles)
        self.assertEqual(len(proba), 3)

    def test_error_handling(self) -> None:
        """Test the error handling in the pipeline."""
        error_filter = ErrorFilter(filter_everything=True)
        filter_reinserter = FilterReinserter.from_error_filter(
            error_filter,
            fill_value=np.nan,
        )
        pipeline = Pipeline(
            steps=[
                ("smiles2mol", SmilesToMol()),
                ("mol2chemprop", MolToChemprop()),
                ("error_filter", error_filter),
                ("model", get_model()),
                ("filter_reinserter", PostPredictionWrapper(filter_reinserter)),
            ],
        )
        smiles = ["CCO", "CNC", "CCN", "invalid"]
        pred = pipeline.predict(smiles)
        self.assertEqual(len(pred), 4)
        self.assertTrue(np.isnan(pred[-1]))
        proba = pipeline.predict_proba(smiles)
        self.assertEqual(len(proba), 4)
        self.assertTrue(np.isnan(proba[-1]).all())

    def test_state_dict_forwarding(self) -> None:
        """Test that the state_dict is properly forwarded to the model."""
        chemprop_classifier = get_classification_pipeline()
        chemprop_classifier.fit(["CCO", "CCN"], [0, 1])
        chemprop_model = chemprop_classifier.named_steps["model"]
        state_dict = chemprop_model.model.state_dict()

        new_chemprop_classifier = get_classification_pipeline()
        new_chemprop_classifier.set_params(
            model__model__state_dict_ref=state_dict,
        )

        orig_pred = chemprop_classifier.predict_proba(["CCO", "CCN"])
        new_pred = new_chemprop_classifier.predict_proba(["CCO", "CCN"])

        self.assertTrue(np.allclose(orig_pred, new_pred))

    def test_state_dict_from_url(self) -> None:
        """Test that the state_dict can be loaded from a URL."""
        chemeleon_url = "dummy_chemeleon_url.pt"
        chemeleon_dummy = BondMessagePassing(d_h=2048)
        dummy_state_dict = randomize_state_dict_weights(chemeleon_dummy.state_dict())

        with mock.patch(
            "molpipeline.utils.file_loading.url_file_loading.requests.get",
        ) as mock_get:
            mock_response = mock.Mock()
            buffer = BytesIO()
            torch.save(dummy_state_dict, buffer)
            buffer.seek(0)
            mock_response.content = buffer.read()
            mock_get.return_value = mock_response
            chemprop_classifier = get_classification_pipeline(
                chemprop_kwargs={
                    "model__message_passing__state_dict_ref": URLFileLoader(
                        chemeleon_url,
                    ),
                    "model__message_passing__d_h": 2048,
                    "model__predictor__input_dim": 2048,
                },
            )
            self.assertEqual(mock_get.call_count, 1)

        # Check that the state dict was loaded correctly
        chemprop_mpnn = chemprop_classifier.named_steps["model"].model
        loaded_state_dict = chemprop_mpnn.message_passing.state_dict()
        self.assertEqual(
            dummy_state_dict.keys(),
            loaded_state_dict.keys(),
        )
        for key, value in dummy_state_dict.items():
            self.assertTrue(torch.equal(value, loaded_state_dict[key]))

        # Check that the prediction works
        pred = chemprop_classifier.predict_proba(["CCO", "CCN"])
        self.assertEqual(len(pred), 2)

    def test_state_dict_from_url_serialization(self) -> None:
        """Test that the state_dict set via URLFileLoader can be serialized."""
        chemeleon_url = "https://dummy_chemeleon_url.pt"
        chemeleon_dummy = BondMessagePassing(d_h=2048)
        dummy_state_dict = randomize_state_dict_weights(chemeleon_dummy.state_dict())

        with mock.patch(
            "molpipeline.utils.file_loading.url_file_loading.requests.get",
        ) as mock_get:
            mock_response = mock.Mock()
            buffer = BytesIO()
            torch.save(dummy_state_dict, buffer)
            buffer.seek(0)
            mock_response.content = buffer.read()
            mock_get.return_value = mock_response
            chemprop_classifier = get_classification_pipeline(
                chemprop_kwargs={
                    "model__message_passing__state_dict_ref": URLFileLoader(
                        chemeleon_url,
                    ),
                    "model__message_passing__d_h": 2048,
                    "model__predictor__input_dim": 2048,
                },
            )
            self.assertEqual(mock_get.call_count, 1)

            # Serialize and deserialize the pipeline
            chemprop_classifier_json = recursive_to_json(chemprop_classifier)
            chemprop_classifier_loaded = recursive_from_json(chemprop_classifier_json)
            self.assertEqual(mock_get.call_count, 2)
        # Check that the state dict was loaded correctly
        chemprop_mpnn = chemprop_classifier_loaded.named_steps["model"].model
        loaded_state_dict = chemprop_mpnn.message_passing.state_dict()
        self.assertEqual(
            dummy_state_dict.keys(),
            loaded_state_dict.keys(),
        )
        for key, value in dummy_state_dict.items():
            self.assertTrue(torch.equal(value, loaded_state_dict[key]))

    def test_sample_weight_forwarding(self) -> None:
        """Test that the sample weights are properly forwarded to the loss function."""
        smiles = ["CCO", "CCN"] * 32
        y = [0, 1] * 32
        sample_weight = np.ones(64) * 0.5

        bce_loss = BCELoss()
        bce_loss.update = mock.MagicMock(side_effect=bce_loss.update)
        model = get_classification_pipeline(
            chemprop_kwargs={"model__predictor__criterion": bce_loss},
        )
        model.fit(smiles, y, model__sample_weight=sample_weight)
        for call_params in bce_loss.update.call_args_list:
            call_args = call_params[0]
            weight = call_args[3]
            self.assertTrue(np.allclose(weight, sample_weight))

    def test_class_weight_no_sample_weight(self) -> None:
        """Test that class_weight is used correctly."""
        # Use a small dummy dataset with 80% of class 0 and 20% of class 1
        smiles = ["CCO", "CCN", "CCC", "CCCl", "c1ccccc1"] * 4
        y = np.array([0, 0, 0, 0, 1] * 4)
        # Test with class_weight='balanced'
        bce_loss = BCELoss()
        bce_loss.update = mock.MagicMock(side_effect=bce_loss.update)
        model = get_classification_pipeline(
            chemprop_kwargs={
                "class_weight": "balanced",
                "model__predictor__criterion": bce_loss,
            },
        )
        model.fit(smiles, y)
        for call_params in bce_loss.update.call_args_list:
            call_args = call_params[0]
            label = call_args[1].detach().cpu().numpy()
            weight = call_args[3].detach().cpu().numpy()
            # Check that weights correspond to class_weight='balanced'
            neg_label_weight_sum = np.sum(weight[label == 0])
            pos_label_weight_sum = np.sum(weight[label == 1])
            self.assertAlmostEqual(neg_label_weight_sum, pos_label_weight_sum)

    def test_class_weight_with_sample_weight(self) -> None:
        """Test that class_weight and sample_weight are combined correctly."""
        # Use a small dummy dataset with 80% of class 0 and 20% of class 1
        smiles = ["CCO", "CCN", "CCC", "CCCl", "c1ccccc1"] * 4
        y = np.array([0, 0, 0, 0, 1] * 4)
        # The last component has double weight
        sample_weight = np.array([1.0, 1.0, 1.0, 1.0, 2.0] * 4)
        # Test with class_weight='balanced'
        bce_loss = BCELoss()
        bce_loss.update = mock.MagicMock(side_effect=bce_loss.update)
        model = get_classification_pipeline(
            chemprop_kwargs={
                "class_weight": "balanced",
                "model__predictor__criterion": bce_loss,
            },
        )
        model.fit(smiles, y, model__sample_weight=sample_weight)
        for call_params in bce_loss.update.call_args_list:
            call_args = call_params[0]
            label = call_args[1].detach().cpu().numpy()
            weight = call_args[3].detach().cpu().numpy()
            # Compute expected weights
            neg_label_weight_sum = np.sum(weight[label == 0])
            pos_label_weight_sum = np.sum(weight[label == 1])
            self.assertAlmostEqual(neg_label_weight_sum * 2, pos_label_weight_sum)


class TestRegressionPipeline(unittest.TestCase):
    """Test the Chemprop model pipeline for regression."""

    def test_prediction(self) -> None:
        """Test the prediction of the regression model."""
        molecule_net_logd_df = pd.read_csv(
            TEST_DATA_DIR / "molecule_net_logd.tsv.gz",
            sep="\t",
            nrows=100,
        )
        regression_model = get_regression_pipeline()
        regression_model.fit(
            molecule_net_logd_df["smiles"].tolist(),
            molecule_net_logd_df["exp"].to_numpy(),
        )
        pred = regression_model.predict(molecule_net_logd_df["smiles"].tolist())

        self.assertEqual(len(pred), len(molecule_net_logd_df))

        model_copy = joblib_dump_load(regression_model)
        pred_copy = model_copy.predict(molecule_net_logd_df["smiles"].tolist())
        self.assertTrue(np.allclose(pred, pred_copy))

        # Test single prediction, this was causing an error before
        single_mol_pred = regression_model.predict(
            [molecule_net_logd_df["smiles"].iloc[0]],
        )
        self.assertEqual(single_mol_pred.shape, (1,))


class TestMultiRegressionPipeline(unittest.TestCase):
    """Test the Chemprop model pipeline for multi-target/multiple regression."""

    def test_prediction(self) -> None:
        """Test the prediction of the multiple regression model."""
        molecule_net_logd_df = pd.read_csv(
            TEST_DATA_DIR / "molecule_net_logd.tsv.gz",
            sep="\t",
            nrows=100,
        )

        # add another target col with gaussian noise
        rng = np.random.default_rng(seed=123)
        molecule_net_logd_df["exp2"] = molecule_net_logd_df["exp"] + rng.normal(
            loc=0.0,
            scale=1.0,
            size=molecule_net_logd_df.shape[0],
        )
        target_cols = ["exp", "exp2"]

        regression_model = get_regression_pipeline(n_tasks=len(target_cols))
        regression_model.fit(
            molecule_net_logd_df["smiles"],
            molecule_net_logd_df[target_cols].to_numpy(),
        )
        pred = regression_model.predict(molecule_net_logd_df["smiles"])
        self.assertEqual(
            pred.shape,
            (
                len(molecule_net_logd_df),
                len(target_cols),
            ),
        )

        model_copy = joblib_dump_load(regression_model)
        pred_copy = model_copy.predict(molecule_net_logd_df["smiles"])
        self.assertTrue(np.allclose(pred, pred_copy))

        # Test single prediction, this was causing an error before
        single_mol_pred = regression_model.predict(
            [molecule_net_logd_df["smiles"].iloc[0]],
        )
        self.assertEqual(single_mol_pred.shape, (1, len(target_cols)))


class TestClassificationPipeline(unittest.TestCase):
    """Test the Chemprop model pipeline for classification."""

    def setUp(self) -> None:
        """Set up repeated variables."""
        molecule_net_bbbp_df = pd.read_csv(
            TEST_DATA_DIR / "molecule_net_bbbp.tsv.gz",
            sep="\t",
            nrows=100,
        )
        smiles_pipeline = get_smiles_checker_pipeline()
        molecule_net_bbbp_df["smiles"] = smiles_pipeline.transform(
            molecule_net_bbbp_df["smiles"],
        )
        molecule_net_bbbp_df = molecule_net_bbbp_df.dropna(subset=["smiles", "p_np"])

        self.molecule_net_bbbp_df = molecule_net_bbbp_df

    def test_prediction(self) -> None:
        """Test the prediction of the classification model."""
        classification_model = get_classification_pipeline()
        classification_model.fit(
            self.molecule_net_bbbp_df["smiles"].tolist(),
            self.molecule_net_bbbp_df["p_np"].to_numpy(),
        )
        pred = classification_model.predict(
            self.molecule_net_bbbp_df["smiles"].tolist(),
        )
        proba = classification_model.predict_proba(
            self.molecule_net_bbbp_df["smiles"].tolist(),
        )
        self.assertEqual(len(pred), len(self.molecule_net_bbbp_df))
        self.assertEqual(proba.shape[1], 2)
        self.assertEqual(proba.shape[0], len(self.molecule_net_bbbp_df))

        model_copy = joblib_dump_load(classification_model)
        pred_copy = model_copy.predict(self.molecule_net_bbbp_df["smiles"].tolist())
        proba_copy = model_copy.predict_proba(
            self.molecule_net_bbbp_df["smiles"].tolist(),
        )

        nan_indices = np.isnan(pred)
        self.assertListEqual(nan_indices.tolist(), np.isnan(pred_copy).tolist())
        self.assertTrue(np.allclose(pred[~nan_indices], pred_copy[~nan_indices]))

        self.assertEqual(proba.shape, proba_copy.shape)
        self.assertTrue(np.allclose(proba[~nan_indices], proba_copy[~nan_indices]))

        # Test single prediction, this was causing an error before
        single_mol_pred = classification_model.predict(
            [self.molecule_net_bbbp_df["smiles"].iloc[0]],
        )
        self.assertEqual(single_mol_pred.shape, (1,))
        single_mol_proba = classification_model.predict_proba(
            [self.molecule_net_bbbp_df["smiles"].iloc[0]],
        )
        self.assertEqual(single_mol_proba.shape, (1, 2))

    def test_calibrated_classifier(self) -> None:
        """Test if the pipeline can be used with a CalibratedClassifierCV."""
        calibrated_pipeline = CalibratedClassifierCV(
            get_classification_pipeline(),
            cv=2,
            ensemble=True,
            method="isotonic",
        )
        calibrated_pipeline.fit(
            self.molecule_net_bbbp_df["smiles"].tolist(),
            self.molecule_net_bbbp_df["p_np"].to_numpy(),
        )
        predicted_value_array = calibrated_pipeline.predict(
            self.molecule_net_bbbp_df["smiles"].tolist(),
        )
        predicted_proba_array = calibrated_pipeline.predict_proba(
            self.molecule_net_bbbp_df["smiles"].tolist(),
        )
        self.assertIsInstance(predicted_value_array, np.ndarray)
        self.assertIsInstance(predicted_proba_array, np.ndarray)
        self.assertEqual(
            predicted_value_array.shape,
            (len(self.molecule_net_bbbp_df["smiles"].tolist()),),
        )
        self.assertEqual(
            predicted_proba_array.shape,
            (len(self.molecule_net_bbbp_df["smiles"].tolist()), 2),
        )

    def test_homogeneous_ensemble_classifier(self) -> None:
        """Test if the pipeline can be used with a HomogeneousEnsembleClassifier.

        To ensure compatibility with HomogeneousEnsembleClassifier classes which use
        joblib for parallelization.

        """
        ensemble_pipeline = HomogeneousEnsembleClassifier(
            estimator=get_classification_pipeline(),
            sampler=2,
        )
        ensemble_pipeline.fit(
            self.molecule_net_bbbp_df["smiles"].tolist(),
            self.molecule_net_bbbp_df["p_np"].to_numpy(),
        )
        predicted_value_array = ensemble_pipeline.predict(
            self.molecule_net_bbbp_df["smiles"].tolist(),
        )
        self.assertIsInstance(predicted_value_array, np.ndarray)
        self.assertEqual(
            predicted_value_array.shape,
            (len(self.molecule_net_bbbp_df["smiles"].tolist()),),
        )


class TestMulticlassClassificationPipeline(unittest.TestCase):
    """Test the Chemprop model pipeline for multiclass classification."""

    def test_prediction(self) -> None:
        """Test the prediction of the multiclass classification model."""
        test_data_df = pd.read_csv(
            TEST_DATA_DIR / "multiclass_mock.tsv",
            sep="\t",
            index_col=False,
        )
        classification_model = get_multiclass_classification_pipeline(n_classes=3)
        mols = test_data_df["Molecule"].tolist()
        classification_model.fit(
            mols,
            test_data_df["Label"].to_numpy(),
        )
        pred = classification_model.predict(mols)
        proba = classification_model.predict_proba(mols)
        self.assertEqual(len(pred), len(test_data_df))
        self.assertEqual(proba.shape[1], 3)
        self.assertEqual(proba.shape[0], len(test_data_df))

        model_copy = joblib_dump_load(classification_model)
        pred_copy = model_copy.predict(mols)
        proba_copy = model_copy.predict_proba(mols)

        nan_mask = np.isnan(pred)
        self.assertListEqual(nan_mask.tolist(), np.isnan(pred_copy).tolist())
        self.assertTrue(np.allclose(pred[~nan_mask], pred_copy[~nan_mask]))

        self.assertEqual(proba.shape, proba_copy.shape)
        self.assertEqual(pred.shape, pred_copy.shape)
        self.assertTrue(np.allclose(proba[~nan_mask], proba_copy[~nan_mask]))

        # Test single prediction, this was causing an error before
        single_mol_pred = classification_model.predict(
            [test_data_df["Molecule"].iloc[0]],
        )
        self.assertEqual(single_mol_pred.shape, (1,))
        single_mol_proba = classification_model.predict_proba(
            [test_data_df["Molecule"].iloc[0]],
        )
        self.assertEqual(single_mol_proba.shape, (1, 3))

        with self.assertRaises(ValueError):
            classification_model.fit(
                mols,
                test_data_df["Label"].add(1).to_numpy(),
            )
        with self.assertRaises(ValueError):
            classification_model = get_multiclass_classification_pipeline(n_classes=2)
            classification_model.fit(
                mols,
                test_data_df["Label"].to_numpy(),
            )
